#include <stdio.h>
#include <stdlib.h>

#define ROWS 512
#define NASC 256
#define COLS 7
#define NIL  -1
#define MAXH 256  /* maximum huffman code length of char */

#define CHAR 0
#define FREQ 1
#define TYPE 2
#define LEFT 3
#define RIGH 4
#define MARK 5
#define LENG 6

#define DEBUG 0

/* binary Huffman tree */
typedef struct {
  char ch;
  float freq;
  char type;  /* T = terminal, N = non-terminal */
  int left;
  int right;
  int marked;
  char huf[MAXH];
  int len;
} CharTab ;

/* print char if printable, ? else */
void print_char(int c) {
  if (c=='\n') printf("NL");
  else if (c==' ') printf("SP");
  else if (c=='\t') printf("\\t");
  else if ((c>=32)&&(c<=126)) printf(" %c",c);
  else printf (" ?");
}

/* find two sub-trees with minimum freq */
int min_freq(CharTab *m, int size, int *a, int *b) {
  int i;
  float f;

  *a=-1;
  *b=-1;
  f=2;
  for (i=0;i<size;i++) {
    if (m[i].freq>0 && m[i].freq<f && !m[i].marked) {
      *a=i;
      f=m[i].freq;
    }
  }
  f=2;
  *b=*a;
  for (i=0;i<size;i++) {
    if (m[i].freq>0 && m[i].freq<f && !m[i].marked && i!=*a) {
      *b=i;
      f=m[i].freq;
    }
  }
  if (DEBUG)
  printf("min_freq: choose a=%d(%c,f=%.3f),b=%d(%c,f=%.3f)\n",
    *a,m[*a].ch,m[*a].freq,*b,m[*b].ch,m[*b].freq);
  if (*b!=*a) return 1;
  else return 0;
}

/* traverse tree and enter huffman character codes in terminals */
void trav(CharTab *m, int entry, char *sofar) {
  char *sl,*sr;
  
  if (m[entry].type=='T') {
    strcpy(m[entry].huf,sofar);
    m[entry].len=strlen(sofar);
  } else {
    sl=(char*)malloc(strlen(sofar)+1);
    sr=(char*)malloc(strlen(sofar)+1);
    strcpy(sl,sofar);
    strcpy(sr,sofar);
    strcat(sl,"0");
    strcat(sr,"1");
    trav(m,m[entry].left,sl);
    trav(m,m[entry].right,sr);
    free(sl);
    free(sr);
  }
}

main(int argc, char *argv[]) {
  FILE *fp;
  int i,j,c,cnt=0,pc=0,a,b,rc,nxt,chars_ori=0,bits_compr=0;
  float avg,elen;
  CharTab m[ROWS];

  printf("<pre>\n");
  for (i=0;i<ROWS;i++) {
    m[i].ch=i;
    m[i].freq=0.0;
    m[i].type='T';
    m[i].left=NIL;
    m[i].right=NIL;
    m[i].marked=0;
    m[i].huf[0]='\0';
  }
  fp=fopen(argv[1],"r");
  while ((c=getc(fp))!=EOF) {
    m[c].freq+=1;
    chars_ori++;
  }
  fclose(fp);
  for (i=0;i<ROWS;i++) {
    m[i].freq/=chars_ori;
  }
  printf("%d characters in source file.\n",chars_ori);
  printf("\nFrequencies:\n\n");
  for (i=0;i<ROWS;i++) {
    if (m[i].freq!=0) {
      pc++;
      print_char(i);
      printf(" %.3f ",m[i].freq);
      if (pc!=0&&pc%8==0) printf("\n");
    }
  }
  printf("\n");
  a=NIL;
  b=NIL;
  rc=1;
  nxt=NASC-1;
  while (rc) {
    rc=min_freq(m,ROWS,&a,&b);
    if (a==b) { break; }
    nxt++; /* new subtree */
    m[a].marked=1;
    m[b].marked=1;
    m[nxt].ch='#';
    m[nxt].freq=m[a].freq+m[b].freq;
    m[nxt].type='N';
    m[nxt].left=a;
    m[nxt].right=b;
    m[nxt].marked=0;
    if (DEBUG)
    printf("main: create %d(f=%.3f) from %d,%d \n",nxt,m[nxt].freq,a,b);
  }
  trav(m,a,"");
  printf("\nHuffman character codes:\n\n");
  avg=0;
  elen=0;
  cnt=0;
  for (i=0;i<ROWS;i++) {
    if (m[i].type=='T' && m[i].freq>0) {
      print_char(i);
      printf(" -> %s\n",m[i].huf);
      avg+=strlen(m[i].huf);
      elen+=strlen(m[i].huf)*m[i].freq;
      cnt++;
    }
  }
  printf("\nAverage code length: %.3f\n\n",avg/cnt);
  printf("\nExpected code length: %.3f\n\n",elen);
  printf("\nSource file in Huffman code (blanks between character encodings):\n\n");
  printf("</pre>\n\n");
  fp=fopen(argv[1],"r");
  while ((c=getc(fp))!=EOF) {
    printf("%s ",m[c].huf);
    bits_compr+=m[c].len;
  }
  printf("\n\n<pre>\nCompressed into %d bits.\n",bits_compr);
  printf("Assuming 8 bit per character that's %.0f%% of original size.\n",
    100*(bits_compr/8.0)/chars_ori);
}

